@@ -79,9 +79,31 @@ setMethod("relistToClass", "XString",
 setMethod("width", "character",
     function(x)
     {
-        if (any(is.na(x)))
+        if (anyNA(x))
             stop("NAs in 'x' are not supported")
-        nchar(x, type="bytes")
+
+        ## Bytes with values > 127 in 'x' break 'nchar(x, type="chars")'
+        ## on some systems, depending on how LC_CTYPE is set. For example
+        ## on my Linux laptop where LC_CTYPE is set to en_US.UTF-8:
+        ##   Sys.getlocale("LC_CTYPE")  # en_US.UTF-8
+        ##   x <- rawToChar(as.raw(135L))
+        ##   nchar(x, type="chars")  # invalid multibyte string, element 1
+        ## These byte values are legit e.g. they occur when doing things
+        ## like 'SolexaQuality(99L)' or 'BString(rawToChar(as.raw(135L)))'.
+        ## The purpose of temporarily setting LC_CTYPE to C below is to
+        ## make sure that the call to 'nchar(x, type="chars")' will always
+        ## succeed and do the right thing. A simple test is:
+        ##   x <- safeExplode(rawToChar(as.raw(1:255)))
+        ##   stopifnot(identical(width(x), rep.int(1L, 255)))
+        ## Note that using 'nchar(x, type="bytes")' wouldn't be an option
+        ## because it breaks the Modstrings package (which uses multibyte
+        ## characters).
+
+        prev_LC_CTYPE <- Sys.getlocale("LC_CTYPE")
+        Sys.setlocale("LC_CTYPE", "C")
+        on.exit(Sys.setlocale("LC_CTYPE", prev_LC_CTYPE))
+
+        nchar(x, type="chars")
     }
 )
 
@@ -154,6 +176,58 @@ setReplaceMethod("subseq", "XStringSet",
 )
 
 
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### make_XStringSet_from_strings()
+###
+### Low-level generic called by XStringSet() constructor. Not intended to be
+### called directly by the end user.
+### Purpose is to make it easy to extend the XStringSet() constructor to
+### support XStringSet derivatives defined in other packages. For example,
+### defining the following method in the Modstrings package will make calls
+### of the form 'XStringSet("ModDNA", ...)' work (granted that seqtype()
+### works properly on ModDNAStringSet objects via appropriate methods):
+###
+###   setMethod("make_XStringSet_from_strings", "ModStringSet",
+###       function(x0, strings, start, width)
+###       {
+###           codec <- modscodec(seqtype(x0))
+###           strings <- vapply(strings,
+###               function(string)
+###                   .convert_letters_to_one_byte_codes(string, codec),
+###               character(1),
+###               USE.NAMES=FALSE)
+###           callNextMethod()
+###       }
+###   )
+###
+
+setGeneric("make_XStringSet_from_strings", signature="x0",
+    function(x0, strings, start, width)
+    {
+        ## Only light checking of 'start' and 'width' (i.e. we don't check
+        ## that they have the same length as 'strings' and define valid
+        ## ranges on its elements).
+        stopifnot(is(x0, "XStringSet"), is.integer(start), is.integer(width))
+        if (!is.character(strings))
+            stop(wmsg("input must be a character vector"))
+        ans <- standardGeneric("make_XStringSet_from_strings")
+        stopifnot(class(ans) == class(x0))
+        ans
+    }
+)
+
+### Default method.
+setMethod("make_XStringSet_from_strings", "XStringSet",
+    function(x0, strings, start, width)
+    {
+        lkup <- get_seqtype_conversion_lookup("B", seqtype(x0))
+        .Call2("new_XStringSet_from_CHARACTER",
+               class(x0), elementType(x0), strings, start, width, lkup,
+               PACKAGE="Biostrings")
+    }
+)
+
+
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### The XStringSet() constructor. NOT exported.
 ###
@@ -161,6 +235,11 @@ setReplaceMethod("subseq", "XStringSet",
 ### Start/End/Width) interface.
 ###
 
+setGeneric("XStringSet", signature="x",
+    function(seqtype, x, start=NA, end=NA, width=NA, use.names=TRUE)
+        standardGeneric("XStringSet")
+)
+
 ### 'x' must be a character string or an XString object.
 .oneSeqToXStringSet <- function(seqtype, x, start, end, width, use.names)
 {
@@ -180,32 +259,25 @@ setReplaceMethod("subseq", "XStringSet",
     extractList(ans_xvector, ans_ranges)
 }
 
-.charToXStringSet <- function(seqtype, x, start, end, width, use.names)
+.charToXStringSet <- function(seqtype, strings, start, end, width, use.names)
 {
-    if (length(x) == 1L) {
-        ans <- .oneSeqToXStringSet(seqtype, x, start, end, width, use.names)
-        return(ans)
+    if (length(strings) == 1L) {
+        ans <- .oneSeqToXStringSet(seqtype, strings,
+                                   start, end, width, use.names)
+    } else {
+        use.names <- normargUseNames(use.names)
+        x0 <- new2(paste0(seqtype, "StringSet"), check=FALSE)
+        solved_SEW <- solveUserSEW(width(strings),
+                                   start=start, end=end, width=width)
+        ans <- make_XStringSet_from_strings(x0, strings,
+                                            start(solved_SEW),
+                                            width(solved_SEW))
+        if (use.names)
+            names(ans) <- names(strings)
     }
-    use.names <- normargUseNames(use.names)
-    ans_elementType <- paste(seqtype, "String", sep="")
-    ans_class <- paste(ans_elementType, "Set", sep="")
-    solved_SEW <- solveUserSEW(width(x), start=start, end=end, width=width)
-    ans <- .Call2("new_XStringSet_from_CHARACTER",
-                 ans_class, ans_elementType,
-                 x, start(solved_SEW), width(solved_SEW),
-                 get_seqtype_conversion_lookup("B", seqtype),
-                 PACKAGE="Biostrings")
-    if (use.names)
-        names(ans) <- names(x)
     ans
 }
 
-
-setGeneric("XStringSet", signature="x",
-    function(seqtype, x, start=NA, end=NA, width=NA, use.names=TRUE)
-        standardGeneric("XStringSet")
-)
-
 setMethod("XStringSet", "character",
     function(seqtype, x, start=NA, end=NA, width=NA, use.names=TRUE)
     {
